from dataclasses import dataclass
from datetime import datetime


@dataclass
class Token:
    value: str
    expire_in: datetime
    refresh_value: str | None = None


@dataclass
class Income:
    id: str
    approved_receipt_uuid: str
    json_url: str
    print_url: str

    data: dict | None = None



from typing import TypeVar, Optional, Union
from datetime import datetime
import random
import string
import time

from moy_nalog.methods import BaseMethod, AuthMethod, AddIncomeMethod
from moy_nalog.http import HttpConnection
from moy_nalog.types import Token

T = TypeVar("T")


class MoyNalog:
    def __init__(self, login: str, password: str) -> None:
        self._login: str = login
        self.__password: str = password

        self._device_id: str = self._create_device_id()

        self._connection: HttpConnection = self._init_http()

        self._auth: AuthMethod = self._init_auth_method()
        self._income: AddIncomeMethod = self._init_income_method()

        self._authorized: bool = False

        self.__token: Optional[Token] = None
        self.__inn: Optional[str] = None

    def _init_http(self) -> HttpConnection:
        return HttpConnection()

    def _init_auth_method(self) -> AuthMethod:
        return AuthMethod(
            self._connection, self._login, self.__password, self._device_id
        )

    def _init_income_method(self) -> AddIncomeMethod:
        return AddIncomeMethod(self._connection)

    def _create_device_id(self) -> str:
        return "".join(random.choices(string.ascii_lowercase + string.digits, k=22))

    async def _make_auth(self) -> dict:
        inn, token = await self._auth.execute()
        self.__inn = inn
        self.__token = token

        self._authorized = True

    async def _get_token(self) -> str:
        if (
            self.__token.value
            and self.__token.expire_in
            and int(time.time() * 1000) + 60 * 1000
            < int(self.__token.expire_in.timestamp() * 1000)
        ):
            return self.__token.value

        if not self._authorized:
            await self._make_auth()

        self.__token = await self._auth.get_token_by_refresh_token(
            connection=self._connection, refresh_token=self.__token.refresh_value
        )

        return self.__token.value

    async def _call(self) -> dict:
        pass

    async def add_income(
        self,
        name: str,
        amount: Union[int, float],
        quantity: int = 1,
        date: str | datetime = datetime.now(),
    ):
        await self._execute_method(
            self._income.execute(),
            inn=self.__inn,
            name=name,
            amount=amount,
            quantity=quantity,
            date=date,
        )

    async def get_user_info(self):
        pass

    async def _execute(self, method: T, **kwargs) -> T:
        return await method.execute(kwargs)

    async def _execute_method(self, method: BaseMethod, **kwargs) -> BaseMethod:
        token = await self._get_token()
        await self._execute(method, token, kwargs)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass

    def __repr__(self) -> str:
        return "MoyNalog()"



from httpx import AsyncClient

BASE_URL = "https://lknpd.nalog.ru/api/v1"

HEADERS = {
    "accept": "application/json, text/plain, */*",
    "accept-language": "ru-RU,ru;q=0.9,en-US;q=0.8,en;q=0.7",
    "content-type": "application/json",
}


class HttpConnection:
    def __init__(self) -> None:
        self._async_client: AsyncClient | None = None

    async def __aenter__(self) -> AsyncClient:
        if not self._async_client:
            self._async_client = AsyncClient(base_url=BASE_URL, headers=HEADERS)
        return self._async_client

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self._async_client.aclose()
        self._async_client = None



import datetime
from decimal import Decimal, ROUND_HALF_UP
from typing import Tuple, Optional, Union
from dataclasses import dataclass

from moy_nalog.http import HttpConnection, HEADERS, BASE_URL
from moy_nalog.exceptions import RefreshTokenNotFoundError, RejectedIncomeError
from moy_nalog.types import Token, Income


@dataclass
class BaseMethod:
    async def _make_request(
        self, connection: HttpConnection, url: str, **kwargs
    ) -> dict:
        async with connection as conn:
            response = await conn.post(url, **kwargs)
            return response.json()

    async def execute(self, connection: HttpConnection):
        pass


@dataclass
class AuthMethod(BaseMethod):
    connection: HttpConnection

    login: str
    password: str
    device_id: str

    _common_body: Optional[dict] = None

    def __post_init__(self):
        self._common_body = {
            "deviceInfo": {
                "sourceDeviceId": self.device_id,
                "sourceType": "WEB",
                "appVersion": "1.0.0",
                "metaDetails": {
                    "userAgent": (
                        "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_2_2)"
                        " AppleWebKit/537.36 (KHTML, like Gecko)"
                        " Chrome/88.0.4324.192 Safari/537.36"
                    )
                },
            }
        }

    async def get_token_by_refresh_token(
        self, connection: HttpConnection, refresh_token
    ):
        body = {**self._common_body, "refreshToken": refresh_token}
        response = await self._make_request(connection, url="/auth/token", json=body)
        return Token(
            value=response.get("Token"),
            expire_in=response.get("tokenExpireIn"),
            refresh_value=response.get("refreshToken")
            if response.get("refreshToken")
            else None,
        )

    async def execute(
        self,
    ) -> Tuple[str, Token]:
        body = {"username": self.login, "password": self.password, **self._common_body}

        response = await self._make_request(
            self.connection, url="/auth/lkfl", json=body
        )
        if response.get("refreshToken") is None:
            raise RefreshTokenNotFoundError(
                f"{response.get("message") or "Cannot get refreshToken"}"
            )
        return response.get("profile").get("inn"), Token(
            value=response.get("token"),
            expire_in=response.get("tokenExpireIn"),
            refresh_value=response.get("refreshToken"),
        )


@dataclass
class AddIncomeMethod(BaseMethod):
    connection: HttpConnection

    def _date_to_local_iso(
        date: datetime.datetime | str = datetime.datetime.now(),
    ) -> str:
        if isinstance(date, str):
            date = datetime.fromisoformat(date)

        offset = date.utcoffset().total_seconds() / 60 if date.utcoffset() else 0
        absoff = abs(offset)

        local_time = date - datetime.timedelta(minutes=offset)

        iso_string = local_time.isoformat(timespec="seconds")

        sign = "-" if offset > 0 else "+"
        hours = int(absoff // 60)
        minutes = int(absoff % 60)

        offset_string = f"{sign}{str(hours).zfill(2)}:{str(minutes).zfill(2)}"

        return iso_string + offset_string

    async def send_income(
        self,
        token: str,
        name: str,
        amount: Union[int, float],
        quantity: int = 1,
        date: datetime.datetime | str = datetime.datetime.now(),
    ):
        amount = Decimal(f"{amount}")
        quantity = Decimal(f"{quantity}")

        headers = {"authorization": f"Bearer {token}", **HEADERS}
        body = {
            "paymentType": "CASH",
            "ignoreMaxTotalIncomeRestriction": False,
            "client": {
                "contactPhone": None,
                "displayName": None,
                "incomeType": "FROM_INDIVIDUAL",
                "inn": None,
            },
            "requestTime": self._date_to_local_iso(),
            "operationTime": self._date_to_local_iso(date),
            "services": [
                {
                    "name": name,
                    "amount": float(amount.quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)),
                    "quantity": int(quantity),
                }
            ],
            "totalAmount": float((amount * quantity).quantize(
                Decimal("0.01"), rounding=ROUND_HALF_UP
            )),
        }
        response = await self._make_request(
            connection=self.connection, url="income", headers=headers, json=body
        )
        return response

    async def get_information_about_income(self, url: str) -> dict:
        return await self._make_request(self.connection, url=url)

    async def execute(
        self,
        inn: str,
        token: str,
        name: str,
        amount: Union[int, float],
        quantity: int = 1,
        date: datetime.datetime | str = datetime.datetime.now(),
    ) -> dict:
        response = await self.send_income(
            token=token, name=name, amount=amount, quantity=quantity, date=date
        )
        if not response or not response.get("approvedReceiptUuid"):
            raise RejectedIncomeError("Check your params.")

        url = f"{BASE_URL}/receipt/{self.inn}/{response.get("approvedReceiptUuid")}"

        income = Income(
            id=response.get("approvedReceiptUuid"),
            approved_receipt_uuid=response.get("approvedReceiptUuid"),
            json_url=f"{url}/json",
            print_url=f"{url}/print",
        )

        return await self.get_information_about_income(income.json_url)
